(* generated by dune *)

(** @canonical Configurator.Dune_lang *)
module Dune_lang = Configurator__Dune_lang

(** @canonical Configurator.Extract_obj *)
module Extract_obj = Configurator__Extract_obj

(** @canonical Configurator.Import *)
module Import = Configurator__Import

(** @canonical Configurator.Ocaml_config *)
module Ocaml_config = Configurator__Ocaml_config

(** @canonical Configurator.V1 *)
module V1 = Configurator__V1

module Configurator__ = struct end
[@@deprecated "this module is shadowed"]
